@@ -1,10 +1,11 @@
 from . import config
+import dask
 from dask.config import config
 from .actor import Actor, ActorFuture
-from .core import connect, rpc
+from .core import connect, rpc, Status
 from .deploy import LocalCluster, Adaptive, SpecCluster, SSHCluster
 from .diagnostics.progressbar import progress
-from .diagnostics.plugin import WorkerPlugin, SchedulerPlugin
+from .diagnostics.plugin import WorkerPlugin, SchedulerPlugin, PipInstall
 from .client import (
     Client,
     Executor,
@@ -17,6 +18,7 @@ from .client import (
     futures_of,
     get_task_stream,
     performance_report,
+    get_task_metadata,
 )
 from .lock import Lock
 from .nanny import Nanny
@@ -38,3 +40,15 @@ versions = get_versions()
 __version__ = versions["version"]
 __git_revision__ = versions["full-revisionid"]
 del get_versions, versions
+
+if dask.config.get("distributed.admin.event-loop") in ("asyncio", "tornado"):
+    pass
+elif dask.config.get("distributed.admin.event-loop") == "uvloop":
+    import uvloop
+
+    uvloop.install()
+else:
+    raise ValueError(
+        "Expected distributed.admin.event-loop to be in ('asyncio', 'tornado', 'uvloop'), got %s"
+        % dask.config.get("distributed.admin.event-loop")
+    )
